#!/usr/bin/env python

#
# Copyright 2012 Frank Adelstein.  All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this
# software is hereby granted without fee, provided that
# the copyright notice and permission notice are not removed.
#
# Last changed: 2013-11-11 by FNA
#   added hist_spell_if_moved() functionality
# Last changed: 2013-11-08 by FNA
#   added end_rubberband_special() functionality
# Last changed: 2012-02-02 by FNA
#


import pygtk
pygtk.require('2.0')
import gtk

import string

import bananaconst
import bananatile

import bananahistory
import bananaevent

import subprocess

#
# FSM_table format: [rec1, rec2, rec3...]
# where recn = [state, event_type, [flags,], [actions,], next_state]
# legal "flags" and "actions" are defined below in code (note that it
# must be a list, so singleton entries need to have a trailing comma
# beofre the close square bracket). 
#
#

class FSM:

    # instance variables
    state = 0
    start = 0
    table = {}
    warn = 0

    flags = ('on_tile', 'not_on_tile', 'this_tile_selected',
             'not_this_tile_selected', 'none_selected', 'one_selected',
             'plural_selected', 'any_selected', )

    actions = ('deselect_all', 'update_position', 'select_tile', 'move_tile',
               'start_rubberband', 'rubberband', 'end_rubberband', 
               'end_rubberband_special', 'hist_spell_if_moved',
               'select_tiles', 'set_move_hist', 'redraw', 'save_position',
               'update_position', 'fix_position', 'check_spelling')

    def __init__(self, FSM_table, start=0, warn=None):
        self.state = start
        self.warn = warn
        self.start = start

        # store the table in an array table indexed by start state
        # (i.e., table[0] = [rec1, rec2, ...] where rec1.state = rec2.state =
        # recn.state = 0)
        for record in FSM_table:
            start_state = record[0]
            if start_state in self.table:
                self.table[start_state].append(record)
            else:
                self.table[start_state] = [record]

    def destroy():
        pass

    # put the FSM back into its start state
    def reset(self):
        self.state = self.start

    #
    # returns True on success, False if problems
    #
    def transition(self, banana, event_name, event):

        # check if there are transitions for current state
        if self.state not in self.table:
            if self.warn:
                msg = "No information for state {}, skipping."
                print msg.format(self.state)
            return False

        # iterate through possible transitions
        for record in self.table[self.state]:

            skip_record = False
            if event_name == record[1]:
                if record[2] != None:
                    # only check flags if some were specified
                    for flag in record[2]:
                        if not self._evaluate_flag(banana, flag, event):
                            if banana.debug:
                                print "flag [{}] is false".format(flag)
                            skip_record = True
                            break
                        else:
                            if banana.debug:
                                print "flag [{}] is true".format(flag)

                    if skip_record:
                        continue

                # matched all the flags
                if record[3] != None:
                    for action in record[3]:
                        if not self._do_action(banana, action, event):
                            # verify the action succeeded
                            # TODO: add a default error state
                            print  "warning: Action [{}] returned False!".format(action)
                        else:
                            if banana.debug:
                                print "Action [{}] succeeded".format(action)

                # set next state
                self.state = record[4]

                if banana.debug:
                    print "state now is in: {}".format(self.state)

                return True
        return False

    def _evaluate_flag(self, banana, flag, event):
        if flag not in self.flags:
            print "Illegal flag [{}] specified".format(flag)
            return False

        # do flag specific check
        if flag == "on_tile":
            return on_tile_flag(banana, event)
        elif flag == "this_tile_selected":
            return this_tile_selected_flag(banana, event)
        elif flag == "not_this_tile_selected":
            return not this_tile_selected_flag(banana, event)
        elif flag == "not_on_tile":
            return not on_tile_flag(banana, event)
        elif flag == "one_selected":
            return (num_selected_flag(banana, event) == 1)
        elif flag == "none_selected":
            return (num_selected_flag(banana, event) == 0)
        elif flag == "plural_selected":
            return (num_selected_flag(banana, event) > 1)
        elif flag == "any_selected":
            return (num_selected_flag(banana, event) > 0)

        return False


    def _do_action(self, banana, action, event):
        if action not in self.actions:
            print "Illegal action [{}] specified".format(action)
            return False

        # do specific action
        if action == "deselect_all":
            return deselect_all_action(banana, event)
        elif action == "select_tile":
            return select_tile_action(banana, event)
        elif action == "redraw":
            return redraw_action(banana, event)
        elif action == "move_tile":
            return move_tile_action(banana, event)
        elif action == "save_position":
            return save_position_action(banana, event)
        elif action == "update_position":
            return update_position_action(banana, event)
        elif action == "fix_position":
            return fix_position_action(banana, event)
        elif action == "start_rubberband":
            return start_rubberband_action(banana, event)
        elif action == "rubberband":
            return rubberband_action(banana, event)
        elif action == "end_rubberband":
            return end_rubberband_action(banana, event)
        elif action == "end_rubberband_special":
            return end_rubberband_action_special(banana, event)
        elif action == "hist_spell_if_moved":
            return hist_spell_if_moved_action(banana, event)
        elif action == "set_move_hist":
            return set_move_hist_action(banana, event)
        elif action == "check_spelling":
            return check_spelling_action(banana, event)

        return False


##################
# flag functions #
##################

# check if cursor is on a tile
def on_tile_flag(banana, event):
    for tile in reversed(banana.tiles):
        if tile.within(event.x, event.y):
            return True
    return False

# checks if the cursor is under a selected tile
def this_tile_selected_flag(banana, event):
    for tile in banana.selected:
        if tile.within(event.x, event.y):
            return True
    return False

# checks if the cursor is under a selected tile
def num_selected_flag(banana, event):
    return len(banana.selected)


####################
# action functions #
####################

# mark all tiles as unselected, delete list
def deselect_all_action(banana, event):
    for tile in banana.tiles:
        tile.selected = 0
    banana.selected = []
    banana.allselected = 0
    return True

# mark top tile under cursor as selected
# saves the new selection event in the undo queue
def select_tile_action(banana, event):

    oldselect = banana.selected
    # eleminiates anything else that had been selected
    deselect_all_action(banana, event)

    for tile in reversed(banana.tiles):
        if tile.within(event.x, event.y):
            tile.selected = 1
            banana.selected=[tile]
            # add to the undo list
            banana.history.do([bananaevent.Select(banana, oldselect, 
                                                  banana.selected)])
            return True
    return False

# redraw the board
def redraw_action(banana, event):
    banana.redraw_cb()
    return True

# call set_move_hist_action() and check_spelling_action()
# only if a tile has actually moved
def hist_spell_if_moved_action(banana, event):

    tile = banana.selected[0]
    if tile.origx == tile.x and tile.origy == tile.y:
        return True

    set_move_hist_action(banana, event)
    check_spelling_action(banana, event)
    return True


# move the single selected tile to current location
def move_tile_action(banana, event):
    # this function only moves 1 tile
    if len(banana.selected) != 1:
        return False

    tile = banana.selected[0]
    # save old position (in case history event needs it)
    tile.origx = tile.x
    tile.origy = tile.y

    if banana.grid:
        # if (snap-to) grid is set, calculate 
        # start of cell = number of tile widths * tilewidth
        tile.x = int(event.x/(banana.tilewidth+1))*(banana.tilewidth+1)
        tile.y = int(event.y/(banana.tilewidth+1))*(banana.tilewidth+1)
    else:
        # just use the cursor location, no snap-to-grid
        tile.x = event.x
        tile.y = event.y

    banana.redraw_cb()
    return True

# save current position of tile and its offset
def save_position_action(banana, event):
    # this function only operates on  1 tile
    #if len(banana.selected) != 1:
    #    return False
    for tile in banana.selected:
        tile.origx = tile.x
        tile.origy = tile.y
        tile.offsetx = tile.x - event.x
        tile.offsety = tile.y - event.y
    return True


# change the current position of tile
def update_position_action(banana, event):
    # this function only operates on  1 tile
    #if len(banana.selected) != 1:
    #    return False
    for tile in banana.selected:
        tile.x = event.x + tile.offsetx
        tile.y = event.y + tile.offsety

    # update image
    banana.redraw_cb()
    return True

# adjust dragged position to line up with grid
def fix_position_action(banana, event):
    # this function only operates on  1 tile
    #if len(banana.selected) != 1:
    #    return False
    if not banana.grid:
        return True

    # Correction is the difference between the current position and
    # the top left start of the cell.  We add half the cell width to
    # "round" the result to the nearest cell.  Correction is the same
    # for all the selected tiles.
    xcorrect = int((banana.selected[0].x + banana.tilewidth/2)/ (banana.tilewidth+1)) * (banana.tilewidth+1) - banana.selected[0].x
    ycorrect = int((banana.selected[0].y + banana.tilewidth/2)/ (banana.tilewidth+1)) * (banana.tilewidth+1) - banana.selected[0].y

    for tile in banana.selected:
        tile.x += xcorrect
        tile.y += ycorrect
    
    # update image
    banana.redraw_cb()
    return True

# save new position of selected items
def set_move_hist_action(banana, event):
    if (not len(banana.selected)):
        return False

    eventlist = []
    for tile in banana.selected:
        eventlist.append(bananaevent.Move(tile, tile.origx, tile.origy, 
                                          tile.x, tile.y))
    banana.history.do(eventlist)
    return True


# save start point of rubberband 
def start_rubberband_action(banana, event):

    banana.startdragx = event.x
    banana.startdragy = event.y
    banana.rubberbanding = True
    return True

# draw rubberband (XOR)
def rubberband_action(banana, event):

    area = banana.labels['area']
    if banana.lastdragx != -1:
        # erase previous rubber band
        makeRectangle(area, banana.xorgc, banana.startdragx,
                      banana.startdragy, banana.lastdragx, 
                      banana.lastdragy)

    # draw new rectangle rubber band
    makeRectangle(area, banana.xorgc, banana.startdragx,
                  banana.startdragy, event.x, event.y)

    # save current position
    banana.lastdragx = event.x
    banana.lastdragy = event.y

    return True

# completete the rubberbanding and select tiles within area
def end_rubberband_action(banana, event):

    oldselect = banana.selected
    # unselect any previous selection
    deselect_all_action(banana, event)
    # erase the previous rubber band if we drew something
    if (banana.lastdragx != -1):
        area = banana.labels['area']
        makeRectangle(area, banana.xorgc, banana.startdragx,
                      banana.startdragy, banana.lastdragx,
                      banana.lastdragy)

    for tile in banana.tiles:
        if tile.inside(banana.startdragx, banana.startdragy, 
                       event.x, event.y):
            # add tile to selected list, if it's within the rectangle
            banana.selected.append(tile)
            tile.selected = 1

    # add to the undo list
    banana.history.do([bananaevent.Select(banana, oldselect, 
                                          banana.selected)])
    # reset status variables
    banana.rubberbanding = 0
    banana.startdragx = -1
    banana.startdragy = -1
    banana.lastdragx = -1
    banana.lastdragy = -1
    return True


# if only one tile was previously selected and the mouse barely 
# moved over empty space (i.e., the user didn't mean to rubberband), 
# then just move the tile, otherwise do the normal end rubberband
def end_rubberband_action_special(banana, event):

    # erase the previous rubber band if we drew something
    if (banana.lastdragx != -1):
        area = banana.labels['area']
        makeRectangle(area, banana.xorgc, banana.startdragx,
                      banana.startdragy, banana.lastdragx,
                      banana.lastdragy)

    # count the number of tiles that would be selected
    count = 0
    for tile in banana.tiles:
        if tile.inside(banana.startdragx, banana.startdragy, 
                       event.x, event.y):
            count += 1

    if (abs(banana.startdragx-event.x) < 5 and
        abs(banana.startdragy-event.y) < 5 and 
        count == 0):
        # call the normal move_tile_action()
        move_tile_action(banana, event)
    else:
        # call the normal end_rubberband()
        end_rubberband_action(banana, event)

    # reset status variables
    banana.rubberbanding = 0
    banana.startdragx = -1
    banana.startdragy = -1
    banana.lastdragx = -1
    banana.lastdragy = -1
    return True

# completete the rubberbanding and select tiles within area
def check_spelling_action(banana, event):

    if banana.allselected or banana.spellmode == 0:
        # skip the spelling check if everything is selected
        # or in no spelling checking mode 
        return True

    for tile in banana.selected:
        # recalculate neighbors 
        bananatile.tileNeighbors(tile, banana.tiles)
    # recalculate words
    words = bananatile.findWords(banana.tiles)
    # print "looking up %d words" % len(words)

    for (t, word, dir) in words:
        # check spelling of each word
        if misspelled(word, banana, banana.debug):
            again = 1
            while (again):
                # mark all misspelled tiles
                t.misspelled = 1

                # navigate to the next tile
                if dir == "down" and t.bottom:
                    t = t.bottom
                elif dir == "right" and t.right:
                    t = t.right
                else:
                    # end look (could just use a break)
                    again = 0
    # redraw so the color changes take effect
    banana.redraw_cb()
    return True


####################################################
# a quick function to change (x1,y1), (x2,y2) into #
# proper (startX, startY, width, height) notation  #
# and call the draw rectangle function             #
####################################################
def makeRectangle (drawable, gc, x1, y1, x2, y2):
    if (x1 < x2):
        x = int(x1)
    else:
        x = int(x2)

    if (y1 < y2):
        y = int(y1)
    else:
        y = int(y2)

    width = int(abs(x1 - x2))
    height = int(abs(y1 - y2))

    drawable.window.draw_rectangle(gc, False, x, y, width, height)

    return

################################
# check the spelling on a word #
################################
def misspelled (word, obj, debug=None):

    # "loose" spelling (the default) means we'll search for the word pattern
    # occurring anywhere in the word list, so just grep for the word

    if obj.spellmode == 2:
        # "strict" spelling means the word MUST be in the list, so
        # change the grep term to force that
        word = '^' + word + '$'

    # use 'grep' to search for word, quiet (no output), case insensitive
    process = subprocess.Popen(["grep", "-qi", word, bananaconst.spellingfile])

    process.wait()
    if process.returncode:
        if debug:
            print "{} is misspelled, rc = {}".format(word, process.returncode)
        return True
    else:
        return False



def main():

    table = (('start', 'click', ('on_tile', 'this_tile_selected'), None, 'clicked'),
       ('start', 
        'click', ('on_tile', 'not_this_tile_selected'), ('select_tile', 'redraw',), 
        'clickedselect'),
       ('start', 'click', ('not_on_tile',), None, 'emptyclick'),

       ('clicked', 
        'release', ('on_tile',), ('deselect_all', 'select_tile', 'redraw'), 
        'start'),
       ('clicked', 'move', 
        None, ('save_position', 'update_position',), 'move'),

       ('clickedselect', 
        'release', ('on_tile',), ('deselect_all', 'select_tile', 'redraw'), 
        'start'),
       ('clickedselect', 'move', 
        None, ('save_position', 'update_position',), 'move'),

       ('emptyclick', 'release', ('one_selected',), ('move_tile', 'set_move_hist', 'check_spelling'),
        'start'),
       ('emptyclick', 
        'release', ('none_selected',), None, 'start'),
       ('emptyclick', 
        'release', ('plural_selected',), ('deselect_all',), 'start'),
       ('emptyclick', 
        'move', ('one_selected',), ('start_rubberband',), 'rubberband'),
       ('emptyclick', 
        'move', ('any_selected',), ('deselect_all', 'start_rubberband'), 
        'rubberband'),
       ('emptyclick', 
        'move', ('none_selected',), ('start_rubberband',), 'rubberband'),

       ('move', 'move', None, ('update_position',), 'move'),
       ('move', 'release', None, ('update_position', 'fix_position', 'set_move_hist', 'check_spelling'), 'start'),

       ('rubberband', 'move', None, ('rubberband',), 'rubberband'),
       ('rubberband', 
        'release', ('one_selected',), ('end_rubberband_special', 'redraw'), 
        'start'),
       ('rubberband', 
        'release', None, ('end_rubberband', 'redraw'), 'start'),

    )

    fsm = FSM(table, 0)

    fsm.transition(None, 'click', None)
    fsm.transition(None, 'release', None)

    fsm.transition(None, 'click', None)
    fsm.transition(None, 'move', None)
    fsm.transition(None, 'move', None)
    fsm.transition(None, 'move', None)
    fsm.transition(None, 'release', None)


if __name__ == "__main__":
    main()
